import os
import sys
import json
import pandas as pd

def tabulate_group(df,list):
    """Creates a tabulate of the given columns 
    Parameters
    ----------
    df: Dataframe we want to rename its columns
    list:List of the columns we want to group by with
    Returns
    -------
    The tabulate
    """
    return(df.groupby(list).size().reset_index())
    
def filter_conditions(df):
    module_name = df['module_name']
    reaction_time = df['reaction_time']

    if module_name in ['FRT-C_Em', 'FRT-C_Em-Pre', 'FRT-C_Em-Post']:
        if reaction_time < 350:
            return "Below"
        elif reaction_time > 2500:
            return "Above"
        else: return "Good"

    elif module_name == 'FRT-C_Mo':
        if reaction_time < 300:
            return "Below"
        elif reaction_time > 2500:
            return "Above"
        else: return "Good"
        
    elif module_name == 'FRT-B_Me':
        if reaction_time < 400:
            return "Below"
        elif reaction_time > 4000:
            return "Above"
        else: return "Good"


def check_files(args):
    if len(args) <= 1:
        sys.exit('Please provide all arguments: \nthe first one is the jatos and \nthe all demographics csv files' )
    else: print(f'You are processing {len(args)-1} demographic files')

    
def upload_jatos(file):
    if not os.path.exists(file):
        sys.exit(f'{file} - does not exist')
    if not os.path.isfile(file):
        sys.exit(f'{file} - is not a file')
    if os.path.splitext(file)[1] == '.txt':     # Load data from the jatos file (txt. file)
        l_d = []
        with open(file, encoding='UTF-8') as f:
            for line in f:
                l_d.append(json.loads(line))

        new_l_d = []
        for item in l_d:
            if isinstance(item, dict):
                new_l_d.append(item)
            elif isinstance(item, int):
                continue
            else:
                new_l_d.extend([b for b in item if b is not None])
        l_d = new_l_d

        for i in range(len(l_d)):
            d = l_d[i]
            if 'responses' in d.keys():
                d.update(json.loads(d['responses']))
                l_d[i] = d
        df_jatos = pd.DataFrame(l_d)
        df_jatos['behavioral_task_duration_dt'] = pd.to_timedelta(df_jatos['behavioral_task_duration'])
        df_jatos['behavioral_task_duration_minutes'] = df_jatos['behavioral_task_duration_dt'].dt.total_seconds() / 60
    else: 
        df_jatos = pd.read_csv(file, low_memory=False)
        df_jatos['participant_id'] = df_jatos['participant_id'].str.strip()
        df_jatos['participant_id'] = df_jatos['participant_id'].astype(str)
        df_jatos['reaction_time'] = pd.to_numeric(df_jatos['reaction_time'], errors= 'coerce')
    
    return df_jatos

def percent_calculation(df):
    if "Above" in df.columns and "Below" in df.columns:
        df["Sum"] = df["Above"].fillna(0) + df["Below"].fillna(0) + df["Good"]
    elif "Above" in df.columns:
        df["Sum"] = df["Above"].fillna(0) + df["Good"]
        print("There are no responses below the minimum.")
    elif "Below" in df.columns:
        df["Sum"] = df["Below"].fillna(0) + df["Good"]
        print ("There are no responses Above the maximum.")
    else:
        df["Sum"] = df["Good"]
        print("There are no responses outside the norm.")
    df["Good_percent"] = df["Good"]/df["Sum"]
    return df


import pandas as pd
from tqdm.auto import tqdm
import time
from fuzzywuzzy import fuzz
from NeuronsMemoryTestPipeline import constants
pd.options.display.show_dimensions = False


def catch_typo(to_correct, presented):
    # short entries are sent to further processing then 
    # Check Levenshtein distance against potential matches

    if len(to_correct) > 3:
        potential_matches = [b for b in presented if fuzz.partial_ratio(to_correct, b) >= 85]
        if len(potential_matches) > 0:
            return True
        # Check common typo patterns against potential matches (add your patterns)
        typo_patterns = {'l': 'i', 'i': 'l', 'p': 'b', 'b': 'p', 'o': 'a', 'a': 'o', 'e': 'i', 'i': 'e'}
        for char, typo in typo_patterns.items():
            if any(typo in to_correct and char in match for match in potential_matches):
                return True
    return False

def catch_abbreviation(to_correct, presented):
    """
    Checks if a phrase matches the abbreviated form of any entry in a given list.
    Args:
    to_correct (str): The phrase to check.
    presented (list of str): The list of full phrases.
    Returns:
    str or None: The original phrase corresponding to the abbreviation if a match is found, otherwise None.
    """
    
    def generate_abbreviation(phrase):
        return ''.join(word[0].upper() for word in phrase.split())
    
    corrected_abbreviation = to_correct.upper()

    for entry in presented:
        if generate_abbreviation(entry) == corrected_abbreviation:
            #print(f'         {to_correct} -> {entry}')
            return entry
    return None

def catch_lower_upper_case(to_correct, presented):
    """
    Checks if a phrase matches the lower or upper case form of any entry in a given list.
    Args:
    to_correct (str): The phrase to check.
    presented (list of str): The list of full phrases.
    Returns:
    str or None: The original phrase corresponding to the abbreviation if a match is found, otherwise None.
    """
    def lower_case(phrase):
        return phrase.lower()

    corrected_abbreviation = to_correct.lower()
    for entry in presented:
        if lower_case(entry) == corrected_abbreviation:
            #print(f'         {to_correct} -> {entry}')
            return entry
    return None


def clean_text_ai(brands_presented, brands_to_clean, model):
    res_all = []
    for brand_to_correct in tqdm(brands_to_clean):
        prompt = constants.SPELLCHECK_PROMPT_TEMPLATE.format(
            potential_responses=brands_presented,
            brand_to_check=brand_to_correct
        )
        try: 
            res = model.generate_content(prompt)
            response_text = res.candidates[0].text.strip()
        except Exception:
            response_text = "Not Found"
        if len(response_text.strip().split()) > 4:
            response_text = "Not Found"
        res_all.append({
            'given_response_label_presented': brand_to_correct,
            'corrected': response_text
        })

        time.sleep(0.3)

    return pd.DataFrame(res_all)



def add_correct_entries(df, brand_cleaned):
    brand_cleaned.columns = ['given_response_label_presented', 'corrected', 'method']
    merged_df = pd.merge(df, brand_cleaned, on='given_response_label_presented', how='left')
    merged_df['recalled_brand'] = merged_df['corrected'].fillna('Not Found')
    merged_df = merged_df.drop(['corrected'], axis=1)
    return merged_df

    
def calculate_corrected_brands(df, brands_presented):
    """
    Calculates the total number of the updated/ corrected brand names obtained from the free text recall
    Args:
        df (pandas df): data frame that contains typed_brand column
    Returns:
        grouped_df: aggregated dataframe per group with counts of corrected brand names.
    """
    count_df = df.loc[df.recalled_brand.isin(brands_presented)].reset_index(drop=True)

    count_df = count_df.drop_duplicates(['participant_id', 'recalled_brand', 'method'])
    corrected_counts = count_df.groupby(['group_id','method']).recalled_brand.value_counts()
    pd.set_option('display.max_rows', None)
    print('*'*50)
    print('\nList of Corrected Entries by Method Used:')
    print(corrected_counts)
    pd.reset_option('display.max_rows')     
    corrected_counts = corrected_counts.reset_index()

    return corrected_counts

def clean_free_recall (df, brands_presented, brands_to_clean, model):
    print('*'*50)
    df['method'] =  ''
    df['corrected'] = ''
    for row in df.itertuples():
        index = row.Index
        brand = row.given_response_label_presented      
        match_abbr = catch_abbreviation(brand, brands_presented)
        match_lower = catch_lower_upper_case(brand, brands_presented)
        if match_abbr:
            #print('ABBREVIATION')
            df.at[index, 'corrected'] = match_abbr
            df.at[index, 'method'] = 'abbreviation'
            
        elif match_lower:
            #print('CASE')
            df.at[index, 'corrected'] = match_lower
            df.at[index, 'method'] = 'case'
        
        else:
            if catch_typo(brand, brands_presented):
                closest_match = sorted(brands_presented, key=lambda b: fuzz.partial_ratio(brand, b), reverse=True)[0]
                #print(f'         {brand} -> {closest_match}')
                #print('FUZZY')
                df.at[index, 'corrected'] = closest_match
                df.at[index, 'method'] = 'fuzzy'
            else:
                if fuzz.ratio(brand, brands_presented) <= 15:
                    df.at[index, 'corrected'] = ''
                    df.at[index, 'method'] = 'not_match'
                else:
                    df.at[index, 'corrected'] = ''
                    df.at[index, 'method'] = 'maybe'
    print(f"\nThere are {df[df.method =='abbreviation'].shape[0]} entries that were indentified as a match using Abbreviation.")
    print(f"\nThere are {df[df.method =='case'].shape[0]} entries that were indentified as a match using Lower Upper Case.")
    print(f"\nThere are {df[df.method =='fuzzy'].shape[0]} entries that were indentified as a match using Fuzzy.")
    df.loc[df['given_response_label_presented'].isin(['others', 'other', 'Others']), 'method'] = 'not_match'
    print(f"\nThere are {df[df.method =='not_match'].shape[0]} entries that do not match targeted Brands at all.")
    print(f"\nThere are {df[df.method =='maybe'].shape[0]} entries that will be further analyzed with AI.")

    further = df[df.method == 'maybe'].copy()
    further['corrected'] = ''
    df = df[df.method != 'maybe'].copy()
    if len(further) > 0:
        clean_df_2 = clean_text_ai(
            brands_presented,
            further['given_response_label_presented'].unique(),
            model=model
        )
        clean_df_2 = clean_df_2[clean_df_2.corrected != 'Not Found'].copy()

        for row in further.itertuples():
            brand = row.given_response_label_presented
            index = row.Index

            if brand in clean_df_2.given_response_label_presented.values:
                try:
                    corrected = clean_df_2.loc[
                        clean_df_2.given_response_label_presented == brand, 'corrected'
                    ].iloc[0]
                    further.at[index, 'corrected'] = corrected
                    further.at[index, 'method'] = 'ai'
                except:
                    further.at[index, 'method'] = 'not_match'
            else:
                further.at[index, 'method'] = 'not_match'

    print(f"\nThere are {further[further.method == 'ai'].shape[0]} entries identified as a match using AI.")

    final_df = pd.concat([df, further], axis=0, ignore_index=True)
    value_counts = final_df['method'].value_counts()
    print('*'*50)
    for method, count in value_counts.items():
        print(f"{method}: {count}")
    return final_df



import pandas as pd
import numpy as np

def produce_demographics(df):
    """ Calculate age, gender, age_group bins demographics for the current 'group_id'
    Args:
        df (dataframe): given jatos dataframe
        id_col (str, optional): name of the column for participant id. Defaults to 'participant_id'.
        age_col (str, optional): name of the column for age data. Defaults to 'Age'.
        age_group (str, optional): name of the column for age_group split data. Defaults to 'Age_Group'.
        gender_col (str, optional): name of the column for gender. Defaults to 'Gender'.
        group_col (str, optional): name of the column(s) for grouping. Defaults to 'group_id'.
    Returns:
        three dataframes: the participants demographics data: age, age split, gender
    """
    age_demographics = df[['group_id','Age']].copy()
    age_demographics= age_demographics.groupby(['group_id'], observed=False)['Age'].agg(['min', 'max', 'mean', 'std']).reset_index()
    age_demographics.rename(columns={'group_id': 'group_id Age'}, inplace=True)
    age_demographics.set_index('group_id Age', inplace=True)
    print(f'AGE Statistics: \n {age_demographics}')

    age_group_count = df[['group_id','Age_Group','participant_id']].drop_duplicates()
    age_group_count= age_group_count.groupby(['group_id', 'Age_Group'], observed=False)['participant_id'].agg(['count']).reset_index()
    age_group_split = df[['group_id','Age_Group','Age']].drop_duplicates()
    age_group_split= age_group_split.groupby(['group_id', 'Age_Group'], observed=False)['Age'].agg(['min', 'max', 'mean', 'std']).reset_index()
    age_group_split = age_group_split.merge(age_group_count, on=['group_id','Age_Group'], how='left')
    print(f'\nAGE GROUP Statistics: \n {age_group_split}')

    df['total_participants'] = df.groupby('group_id')['participant_id'].transform('nunique')
    gender_demographics = df[['group_id', 'Gender', 'participant_id', 'total_participants']].drop_duplicates()
    gender_demographics= gender_demographics.groupby(['group_id', 'Gender', 'total_participants'], observed=False)['participant_id'].agg(['count']).reset_index()
    gender_demographics['percentage'] = np.round(100 * gender_demographics['count']/gender_demographics['total_participants'], 2)
    gender_demographics.reset_index(drop = True, inplace=True)
    gender_demographics.set_index(['group_id', 'total_participants', 'Gender'], inplace=True)
    print(f'\nGENDER Statistics: \n {gender_demographics}')
    return age_demographics, gender_demographics, age_group_split


def filter_reaction_time(df,module, bottom_t = 300, upper_t = 2500, participant_id_column = "participant_id"):
    """ Filter the data by specified upper and lower limits of the reaction time by participant
    Args:
        df (dataframe): : data frame with the results obtained from jatos
        column_name (str): the name of the column that should be considered for filtering
        lt (int, optional): Lower boundary. Defaults to 1000.
        ut (int, optional): Upper boundary. Defaults to 2500.
    Returns:
        dataframe: filtered dataframe
    """
    df = df[(df['module_name'] == module) & ((df['trial_name'] == 'Association'))].copy()
    print(f'Reaction Time for {module}: \n  {df["reaction_time"].describe().to_frame().T}')
    above_reaction_time_t = df.loc[(df.reaction_time > upper_t)]
    print('\nNumber of answers above the threshold:', len(above_reaction_time_t))
    print(round(len(above_reaction_time_t)/(df.shape[0] * 100), 2), '% of the responses')
    below_reaction_time_t = df.loc[(df.reaction_time < bottom_t)]
    print('Number of answers below the threshold:', len(below_reaction_time_t))
    print(round(len(below_reaction_time_t)/(df.shape[0] * 100), 2), '% of the responses')
    df_filtered = df.loc[df.reaction_time.between(bottom_t, upper_t, inclusive="neither")].reset_index(drop = True)
    print("With applying this filter we are excluding", (df[participant_id_column].nunique()-df_filtered[participant_id_column].nunique()), "participants from the initial sample")
    return df_filtered


def merge_dfs(df1, df2, group_columns=['frt_stimulus', 'frt_association', 'module_name', 'group_id']):
    """ Merging function, merge two pandas dataframe based on a list of grouping columns.
    Args:
        df1 : pandas dataframe
        df2 : pandas dataframe
        group_columns : list: Grouping columns, by default == ['frt_stimulus', 'frt_association', 'module_name']
    Returns :
        dataframe: Merged dataframe with column from both input dataframes(df1, df2)
    """
    merged_df = pd.merge(df1, df2, on=group_columns)
    return merged_df 

def process_data(df, ub=2500, group_columns=['frt_stimulus', 'frt_association', 'module_name', 'frt_response', 'group_id']):
    """ Process raw data to generate certainty, mean per participant and a pivot table of response counts: produce total count of groupings + renaming
    Args:
        df (dataframe): pre-filtered database
        ub (int, optional): _description_. Defaults to 25000.
        group_columns (list, optional): Columns to be used for grouping. Defaults to ['frt_stimulus', 'frt_association', 'module_name', 'frt_response', 'group_id'].
    Returns:
        dataframe: pivoted dataframe with total count of Yes and No responses
    """
    df.loc[:, 'certainty'] = ub - df['reaction_time'].copy()
    df.loc[:, 'mean_per_p'] = df.groupby(['participant_id', 'module_name'])['certainty'].transform('mean')

    gr_n = df.groupby(group_columns).size().reset_index(name='count')
    pivot_index = [x for x in group_columns if x != 'frt_response']
    gr_n_p = gr_n.pivot(index=pivot_index, columns='frt_response', values='count')
    gr_n_p = gr_n_p.reset_index()

    gr_n_p = gr_n_p.fillna(0)
    gr_n_p['Yes'] = 0 if 'Yes' not in gr_n_p.columns else gr_n_p['Yes']
    gr_n_p['No'] = 0 if 'No' not in gr_n_p.columns else gr_n_p['No']
    gr_n_p.rename(columns={"No": "N_grouping_No", "Yes": "N_grouping_Yes"}, inplace=True)
    gr_n_p['N_grouping_total'] = gr_n_p['N_grouping_No'] + gr_n_p['N_grouping_Yes']
    return gr_n_p


def group_quartiles(df, group_columns=['frt_stimulus', 'frt_association', 'module_name', 'frt_response', 'group_id'], value_col='certainty', group_zeroes=['participant_id', 'group_id', 'frt_response'] ):
    """ Getting quartiles : IQR
    Args:
        df (dataframe): preprocessed dataframe
        group_columns (list, optional): Columns to be used for grouping. Defaults to ['frt_stimulus', 'frt_association', 'module_name', 'frt_response', 'group_id'].
        value_col (str, optional): The column with the values used to calculate IQR. Defaults to 'certainty'.
        group_zeroes (list, optional): _description_. Defaults to ['participant_id', 'frt_response'].
    Returns:
        dataframe: updated dataframe with the IQR counts per group.
    """
    quartiles = df.groupby(group_columns)[value_col].apply(lambda x: np.percentile(x, [25, 75], interpolation='weibull'))
    quartiles_df = pd.DataFrame(quartiles.tolist(), columns=['IQR1_grouping', 'IQR3_grouping'], index=quartiles.index)
    quartiles_df['group_IQR'] = quartiles_df['IQR3_grouping'] - quartiles_df['IQR1_grouping']
    if (quartiles_df['group_IQR'] < 20).any():
        print("\n ***** There is values in Group IQR smaller than 20! ***** \n")
    result = merge_dfs(quartiles_df, df, group_columns=group_columns)
    result['group_IQR'] = result['group_IQR'].fillna(0)

    result['participant_mean'] = result.groupby(group_zeroes)['group_IQR'].transform('mean')
    result['group_IQR'] = np.where(result['group_IQR']<20, result['participant_mean'], result['group_IQR'])
    result.drop(columns=['participant_mean'], inplace=True)
    return result

def groupby_and_tabulate(df, group_columns=['frt_stimulus', 'frt_association', 'module_name', 'group_id'], value_col='group_IQR'):
    """ Grouping and Calculating Mean, 
        Pivoting to Wide Format, 
        Merging with Original DataFrame, 
        Handling 'Yes' and 'No' Columns, 
        Column Renaming and Dropping Duplicates.
    Args:
        df (dataframe): provided dataframe
        group_columns (list, optional): _description_. Defaults to ['frt_stimulus', 'frt_association', 'module_name', 'group_id'].
        value_col (str, optional): Column to use to calculate mean. Defaults to 'group_IQR'.
    Returns:
        dataframe: updated dataframe
    """
    grouped = df.groupby(group_columns + ['frt_response'])[[value_col]].mean().reset_index()
    pivot = grouped.pivot_table(index=group_columns, columns='frt_response', values=value_col).reset_index()
    merged = pd.merge(df[group_columns], pivot, on=group_columns)
    merged['Yes'] = 0 if 'Yes' not in merged.columns else merged['Yes']
    merged['No'] = 0 if 'No' not in merged.columns else merged['No']
    merged['Yes'] = merged['Yes'].fillna(0)
    merged['No'] = merged['No'].fillna(0)
    merged = merged.rename(columns={'Yes': 'g_iqr_yes', 'No': 'g_iqr_no'})
    merged = merged.drop_duplicates(subset=group_columns)
    return merged

def positive_negative(df, group_columns=['frt_stimulus', 'frt_association', 'module_name', 'group_id']):
    """ Identify if the most responses are Yes or No and create the column with the matching positive or negative label
    Args:
        df (dataframe): working dataframe
        group_columns (list, optional): grouping parameter. Defaults to ['frt_stimulus', 'frt_association', 'module_name', 'group_id'].
    Returns:
        dataframe: updated with the new column positive_negative
    """
    grouped = df.groupby(group_columns)['frt_response'].value_counts().reset_index(name="positive_negative")
    pivot = grouped.pivot_table(index=['frt_association', 'module_name'], columns='frt_response', values="positive_negative").reset_index()
    pivot['Yes'] = 0 if 'Yes' not in pivot.columns else pivot['Yes']
    pivot['No'] = 0 if 'No' not in pivot.columns else pivot['No']
    pivot['Yes'] = pivot['Yes'].fillna(0)
    pivot['No'] = pivot['No'].fillna(0)
    pivot["positive_negative"]=np.where(pivot['Yes']>=pivot["No"],"Positive", "Negative")
    return (pivot)

def calculate_certainty_formula(row, group_columns, iqr_col):
    """ Calculation of the certainty on the group_columns level, using encoded formula.
    Args:
        row (str or int):row of the dataframe to use for calculation
        group_columns (list): list of the name of the columns use for grouping
        iqr_col (str): name of the column use for calculation of the certainty (mainly group_iqr_yes or group_iqr_no)
    Returns:
        float: value that is being calculated
    """
    try: 
        numerator = row['certainty'] * (row[group_columns] / row['N_grouping_total']) * 2
        denominator = row['mean_per_p']
        exponent = ((row['g_iqr_yes'] + row['g_iqr_no']) / row[iqr_col]) / 2
        return (numerator / denominator) ** exponent
    except:
        print ('*** WARNING ***') 
        print(f"group_IQR {row['mean_per_p']} OR grouping size {row['N_grouping_total']} OR exponent {exponent} issue! \n")
        return 0
    
def calculate_certainty_score(row):
    """ Identifies the parameters of the row for further calculation of the certainty score
    Args:
        row (str or int): row of the dataframe
    Returns:
        float: value of the certainty score with the specific parameters
    """
    if row['positive_negative'] == "Positive":
        if row['N_grouping_No'] < 3:
            return 'small_no'
        if row['N_grouping_Yes'] < 3: 
            return 'small_yes'
        if row['frt_response'] == "Yes":
            return calculate_certainty_formula(row, 'N_grouping_Yes', 'g_iqr_yes')
        elif row['frt_response'] == "No":
            return calculate_certainty_formula(row, 'N_grouping_Yes', 'g_iqr_no')
    elif row['positive_negative'] == "Negative":
        if row['N_grouping_No'] < 3:  ######## CHECK and UPDATE
            return 'small_no'
        if row['N_grouping_Yes'] < 3: 
            return 'small_yes'
        if row['frt_response'] == "Yes":
            return calculate_certainty_formula(row, 'N_grouping_No', 'g_iqr_yes')
        elif row['frt_response'] == "No":
            return calculate_certainty_formula(row, 'N_grouping_No', 'g_iqr_no')


def produce_frt_score(df, group_columns=['frt_stimulus', 'frt_association', 'module_name', 'group_id'], certainty_col='certainty_score', response_col='frt_response', certainty='certainty'):
    """ Final step of the frt_score calculation that creates certainty agreement per group, calculates overall % of the yes and no responses and returns the metric frt score.
    Args:
        df (dataframe): dataframe preprocessed with the required columns
        group_columns (list, optional): Columns required for grouping. Defaults to ['frt_stimulus', 'frt_association', 'module_name', 'group_id', 'positive_negative'].
        certainty_col (str, optional): name of the column. Defaults to 'certainty_score'.
        response_col (str, optional): name of the column. Defaults to 'frt_response'.
        certainty (str, optional): name of the column. Defaults to 'certainty'.
    Returns:
        dataframe: the dataframe that contains frt_scores for each stimilus.
    """
    # BREAK CERTAINTY
    df_edge = df[(df[certainty_col] == 'small_no') | (df[certainty_col] == 'small_yes')].copy()
    df = df[(df[certainty_col] != 'small_no') & (df[certainty_col] != 'small_yes')].copy()
    grouped_df = df.groupby(group_columns).agg(
        certainty_sum_Yes = (certainty_col, lambda x: x[df[response_col] == 'Yes'].sum()),
        certainty_sum_No = (certainty_col, lambda x: x[df[response_col] == 'No'].sum()),
        certainty_Yes = (certainty, lambda x: x[df[response_col] == 'Yes'].mean()),
        certainty_No = (certainty, lambda x: x[df[response_col] == 'No'].mean()), 
        agreement_Yes = ('g_iqr_yes', 'mean'),
        agreement_No = ('g_iqr_no', 'mean'),
        N_Yes = (response_col, lambda x: (x == 'Yes').sum()),
        N_No = (response_col, lambda x: (x == 'No').sum()),
        N_Total = (response_col, 'count'),
    ).reset_index()
    
    grouped_df['%_Yes'] = (grouped_df['N_Yes'] / grouped_df['N_Total']) * 100
    grouped_df['%_No'] = (grouped_df['N_No'] / grouped_df['N_Total']) * 100
    grouped_df['frt_summation'] = grouped_df['certainty_sum_Yes'] + grouped_df['certainty_sum_No']
    grouped_df['frt_score'] = (grouped_df['certainty_sum_Yes'] / grouped_df['frt_summation']) * 100
    grouped_df['Scaling'] = (grouped_df['frt_score'] - grouped_df['%_Yes'])
    print('\n***** PLEASE CHECK: distribution of Scaling! *****')
    print(grouped_df['Scaling'].describe().to_frame().T)
    
    if df_edge.shape[0] > 0:
        df_edge['frt_score'] = 0
        df_edge['Scaling'] = 0
        df_edge.loc[df_edge[certainty_col] == 'small_no', 'frt_score'] = 100
        #df_edge.loc[df_edge[certainty_col] == 'small_yes', 'frt_score'] = 0
        #df_edge.loc[df_edge[certainty_col] == 'small_no', 'Scaling'] = 0
        #df_edge.loc[df_edge[certainty_col] == 'small_yes', 'Scaling'] = 0  #### CHECK and UPDATE
        df_edge['N_Yes'] = df_edge['N_grouping_Yes']
        df_edge['N_No'] = df_edge['N_grouping_No']
        df_edge['N_Total'] = df_edge['N_grouping_Yes'] + df_edge['N_grouping_No']
        df_edge['%_Yes'] = (df_edge['N_grouping_Yes'] / df_edge['N_Total']) * 100
        df_edge['%_No'] = (df_edge['N_grouping_No'] / df_edge['N_Total']) * 100
        df_edge = df_edge[group_columns + ['frt_score', 'N_Yes', 'N_No', 'N_Total', '%_Yes', '%_No', 'Scaling']].copy()
        df_edge.drop_duplicates(inplace=True)
        df_edge = df_edge.dropna(axis=1, how='all')
        grouped_df = grouped_df.dropna(axis=1, how='all')  
        grouped_df = pd.concat([grouped_df, df_edge], axis=0)
    return grouped_df

def frt_scores(df):
    """Step-by-step application of the memory recognition computation.
    Args:
        df (dataframe): initial dataframe
    Returns:
        dataframe: Dataframe with required scores including overall % and scores based on IQR and reaction time.
    """
    pre_process_overall = process_data(df)
    merged = merge_dfs(df, pre_process_overall)
    overall_quartiles = group_quartiles(merged)
    overall_quartiles['group_IQR'].describe()
    overall_response_d = groupby_and_tabulate(overall_quartiles)
    merged_iqr_n = merge_dfs(overall_quartiles, overall_response_d)
    pos_neg=positive_negative(merged_iqr_n)
    merged_iqr_n=pos_neg.merge(merged_iqr_n,on=['frt_association', 'module_name'])
    merged_iqr_n['certainty_score']  = merged_iqr_n.apply(calculate_certainty_score, axis=1)
    print(merged_iqr_n[(merged_iqr_n['certainty_score'] != 'small_no') & (merged_iqr_n['certainty_score'] != 'small_yes')]['certainty_score'].astype('float').describe().to_frame().T)
    print('\n***** PLEASE CHECK: edge cases! *****')
    print('Small number of No: ', merged_iqr_n[merged_iqr_n['certainty_score'] == 'small_no']['N_grouping_No'].unique())
    print('Small number of Yes: ', merged_iqr_n[merged_iqr_n['certainty_score'] == 'small_yes']['N_grouping_Yes'].unique())

    overall_frt = produce_frt_score(merged_iqr_n, certainty_col='certainty_score', response_col='frt_response')
    frt_score = overall_frt[['group_id', 'frt_stimulus', 'frt_association','frt_score']].round(2)
    frt_score.sort_values(['group_id','frt_stimulus']).reset_index(drop=True)
    frt_score = frt_score.sort_values(['group_id','frt_score']).reset_index(drop=True)
    return frt_score, overall_frt


def split_score(df, filter_list, column):
    """ Applies the filters, calculates recognition scores for each filtered DataFrame, 
        and merges them based on specific columns and suffixes.
    Args:
        df (DataFrame): DataFrame
        filter_list (list): list of filters: Male or Female,  Younger, Mid, Older
        column (str): a column on which to apply the filters: Gender or Age_Group
    Returns:
        DataFrame: DataFrame with the frt_scores based on the filters
    """
    split = pd.DataFrame()
    for filter_value in filter_list:
        filtered_df = df[df[column] == filter_value]
        split_part, split_part_scaling = frt_scores(filtered_df)
        split_part[column] = filter_value
        split = pd.concat([split, split_part], axis=0)
        split[column] = pd.Categorical(split[column], categories=filter_list, ordered=True)

    split.sort_values(['group_id','frt_stimulus','frt_association']+[column], inplace=True)
    index_cols = ['group_id', 'frt_stimulus', 'frt_association'] + [column]
    split.set_index(index_cols, inplace=True)
    return split, split_part_scaling



# MRT MODULE
MRT_COLUMNS = ['reason_to_end_the_behavioral_task_code', #'participant_quality', 
           'participant_id', 'group_id', 'provider_id', 'module_name',
           'project_identifier', 'reaction_time', 'task_name', 'trial_name','given_response_label_presented', 'procedure_name',
           'trial_stimulus_category','expected_response', 'given_response_accuracy', 'given_response_label_english', 'Gender', 'Age'
          ]
PROJECT = 'neurons-ml'
LOCATION = 'us-central1'

MODEL_TYPE = 'gemini-2.0-flash'
PARAMETERS = {
    "candidate_count": 1,
    "max_output_tokens": 32,
    "temperature": 0.1,
    "top_p": 0.4,
    "top_k": 10
}

# FRT MODULE

FRT_COLUMNS_RENAME = {
    'procedure_stimulus' : 'frt_stimulus',
    'trial_association_english' : 'frt_association',
    'given_response_label_english': 'frt_response'
}

SPELLCHECK_PROMPT_TEMPLATE = """
You are a **Brand Name Spell Checker**.

## Task
Given:
- A list of valid brand names:  
  **potential_responses = {potential_responses}**

- A single input to validate:  
  **brand_to_check = {brand_to_check}**

Determine if `brand_to_check` approximately matches any entry in `potential_responses` using:
- Spelling similarity ≥ 98%
- Edit distance ≤ 2
- Semantic similarity ≥ 90%

## Output Rules
- If a close match exists, return only the **closest valid brand name** from `potential_responses`.
- If no suitable match exists, return exactly: **"Not Found"**


## Examples
- "Addidas" → "Adidas"
- "Nke" → "Nike"
- "ume" → "Puma"
- "Bang" → "Bang & Olufsen"
- "Can't remember" → "Not Found"
- "I don't know" → "Not Found"
- "Life" → "Not Found"

**Return format: ONE brand name from `potential_responses` or "Not Found" — nothing else.**
## Response Template
example: Prada
example: Not Found
"""



import pandas as pd
import numpy as np
from NeuronsMemoryTestPipeline import MRT_text_processing
#from NeuronsMemoryTestPipeline.MRT_text_processing import clean_free_recall
pd.options.display.show_dimensions = False

def alias_group_define(df):
    """obtain all group values, split them and create a new column for each group_id
    Args:
        df (_type_): the alias dataframe
    Returns:
        _type_: updated dataframe with new group_id column
    """
    df.group_id = df.group_id.astype(str)
    new_rows = []
    for _, row in df.iterrows():
        if ',' in row['group_id']:
            groups = row['group_id'].split(',')
            for group in groups:
                new_row = row.copy()
                new_row['group_id_new'] = group
                new_rows.append(new_row)
    new_df = pd.concat([df, pd.DataFrame(new_rows)], ignore_index=True)
    if ',' in row['group_id']:
        new_df.loc[new_df['group_id_new'].isna(), 'group_id_new'] = new_df['group_id']
        new_df['group_id'] = new_df['group_id_new']
        new_df.drop('group_id_new', axis=1, inplace=True)
    new_df.dropna(subset=['group_id'], inplace=True)
    new_df.drop_duplicates(inplace=True)
    new_df = new_df.sort_values('group_id').reset_index(drop=True)
    new_df.group_id = new_df.group_id.astype(str)
    return new_df

def process_recall(df, filter_task, model, alias_df, group=['group_id']):
    """ Prepares the data, process it using text cleaning module, calculates free recall scores per group, including average position and average recall %.
    Args:
        df: dataframe with data from Jatos and identified MRT module plus task_name
        group (list): group by list that provides filter for grouping dataset
    Returns:
        df: updated data frame with the score column
    """
    df['group_id'] = df['group_id'].astype(str)
    alias_df['group_id'] = alias_df['group_id'].astype(str)
    
    df = df.loc[(df['task_name'] == filter_task) & (df['trial_name'].str.startswith('R-'))].copy()
    df['position'] = df['trial_name'].apply(lambda x: int(x.split('-')[-1]))
    
    df_cleaned = pd.DataFrame()
    corrected_entries = pd.DataFrame()
    for group in sorted(df.group_id.unique().tolist(), reverse=False):
        print(f'\n\n *** Group ID: {group} ***')
        df_group = df[df.group_id == group].copy()
        target_brands = (
            alias_df
            .loc[alias_df['group_id'] == group, 'mrt_stimulus']
            .dropna()
            .astype(str)
            .str.strip()
            .loc[lambda x: x != '']
            .unique()
            .tolist()
        )
        print(f'\nBRANDS used in the project for group {group}:')
        for brand in sorted(target_brands):
            print('  ', brand)
        
        total_responses = len(df_group.given_response_label_presented.dropna())
        print('*'*50)
        print(f'\nThere are {total_responses} entries in total.\n')
        print('Identified prior to adjustment: \n', df_group[df_group.given_response_label_presented.isin(target_brands)].given_response_label_presented.value_counts())
        df_group_tocorrect = df_group[~df_group.given_response_label_presented.isin(target_brands)].copy()
        df_group_tocorrect = df_group_tocorrect.dropna(subset=['given_response_label_presented'])
        df_group = df_group[df_group['given_response_label_presented'].isin(target_brands) | pd.isna(df_group['given_response_label_presented'])].copy()
        df_group['corrected'] = df_group['given_response_label_presented']
        df_group['method'] = 'original'
        responses_to_clean = df_group_tocorrect.given_response_label_presented.dropna().unique().tolist()
        print(f'\nThere are {len(responses_to_clean)} entries to clean further.')
        
        corrected_df = MRT_text_processing.clean_free_recall(df_group_tocorrect, target_brands, responses_to_clean, model)
        
        df_group = pd.concat([df_group, corrected_df], axis=0, ignore_index=True)
        
        df_cleaned = pd.concat([df_cleaned, df_group], axis=0, ignore_index=True)
        
    corrected_entries = df_cleaned[['group_id', 'given_response_label_presented', 'corrected', 'method']].copy()
    corrected_entries = corrected_entries[corrected_entries['method'] != 'not_match']
    corrected_entries = corrected_entries.sort_values('method', ascending=True).reset_index(drop=True)
    
    not_match = df_cleaned[['group_id', 'given_response_label_presented', 'corrected', 'method']].copy()
    not_match = not_match[not_match['method'] == 'not_match']
        
    print('*'*50)
    return df_cleaned, corrected_entries, not_match

def compute_scores(df_given: pd.DataFrame, group = ['group_id', 'project_identifier', 'task_name', 'corrected','total_participants']):
    """ Calculates: Average % , Average position, Average Log score based on position
    Args:
        df (dataframe): processed data frame with cleaned text
        no_recall (dataframe): dataset with brands that were not mentioned
    Returns:
        dataframe: updated dataframe with the scores columns
    """
    df = df_given.copy()
    df['corrected'] = df['corrected'].fillna(df['given_response_label_presented'])
    df['position_mean'] = df.groupby(group, observed=False)['position'].transform('mean')
    df['brand_count'] = df.groupby(group, observed=False)['position'].transform('count')   
    df['recall_%'] = df['brand_count'] / df['total_participants'] * 100  
    df['log_position'] = np.log(df['position_mean'] + 1)
    df['Freerecall'] = 1 / np.log(df['position_mean']+1) * df['recall_%']
    df['Freerecall'] = df['Freerecall'].clip(0, 100)
    df = df.sort_values('Freerecall', ascending=False).drop_duplicates().reset_index(drop=True)
    df = df [group + ['position_mean', 'brand_count', 'recall_%', 'log_position', 'Freerecall']].copy()
    df.rename(columns={'corrected':'mrt_stimulus'}, inplace=True)
    df = df.drop_duplicates().reset_index(drop=True)    
    return df